@@ -6,6 +6,9 @@
 
 #include <glib/gi18n-lib.h>
 
+#define GMENU_I_KNOW_THIS_IS_UNSTABLE
+#include <gmenu-desktopappinfo.h>
+
 #include <meta/display.h>
 
 #include "cinnamon-app-private.h"
@@ -57,7 +60,7 @@ struct _CinnamonApp
                           * want (e.g. it will be of TYPE_NORMAL from
                           * the way cinnamon-window-tracker.c works).
                           */
-  GDesktopAppInfo *info;
+  GMenuDesktopAppInfo *info;
 
   CinnamonAppRunningState *running_state;
 
@@ -67,6 +70,7 @@ struct _CinnamonApp
   char *unique_name;
 
   gboolean hidden_as_duplicate;
+  gboolean is_flatpak;
 };
 
 G_DEFINE_TYPE (CinnamonApp, cinnamon_app, G_TYPE_OBJECT);
@@ -108,11 +112,23 @@ cinnamon_app_get_property (GObject    *gobject,
 const char *
 cinnamon_app_get_id (CinnamonApp *app)
 {
-  if (app->entry)
-    return gmenu_tree_entry_get_desktop_file_id (app->entry);
+  if (app->info)
+  {
+    return g_app_info_get_id (G_APP_INFO (app->info));
+  }
   return app->window_id_string;
 }
 
+const char *
+cinnamon_app_get_flatpak_app_id (CinnamonApp *app)
+{
+  if (app->info)
+  {
+    return gmenu_desktopappinfo_get_flatpak_app_id (app->info);
+  }
+  return NULL;
+}
+
 static MetaWindow *
 window_backed_app_get_window (CinnamonApp     *app)
 {
@@ -339,7 +355,7 @@ cinnamon_app_get_keywords (CinnamonApp *app)
     return app->keywords;
 
   if (app->info)
-    keywords = g_desktop_app_info_get_keywords (app->info);
+    keywords = gmenu_desktopappinfo_get_keywords (app->info);
   else
     keywords = NULL;
 
@@ -371,7 +387,8 @@ cinnamon_app_get_nodisplay (CinnamonApp *app)
 
   if (app->entry)
     {
-      return g_desktop_app_info_get_nodisplay (app->info);
+      g_return_val_if_fail (app->info != NULL, TRUE);
+      return gmenu_desktopappinfo_get_nodisplay (app->info);
       // return !g_app_info_should_show (G_APP_INFO (app->info));
     }
 
@@ -651,9 +668,9 @@ cinnamon_app_can_open_new_window (CinnamonApp *app)
     return FALSE;
 
   /* If the app is explicitly telling us, then we know for sure */
-  if (g_desktop_app_info_has_key (G_DESKTOP_APP_INFO (app->info),
+  if (gmenu_desktopappinfo_has_key (GMENU_DESKTOPAPPINFO (app->info),
                                   "X-GNOME-SingleWindow"))
-    return !g_desktop_app_info_get_boolean (G_DESKTOP_APP_INFO (app->info),
+    return !gmenu_desktopappinfo_get_boolean (GMENU_DESKTOPAPPINFO (app->info),
                                             "X-GNOME-SingleWindow");
 
   /* In all other cases, we don't have a reliable source of information
@@ -675,6 +692,18 @@ cinnamon_app_get_state (CinnamonApp *app)
   return app->state;
 }
 
+/**
+ * cinnamon_app_get_is_flatpak:
+ * @app: a #CinnamonApp
+ *
+ * Returns: TRUE if #app is a flatpak app, FALSE if not
+ */
+gboolean
+cinnamon_app_get_is_flatpak (CinnamonApp *app)
+{
+  return app->is_flatpak;
+}
+
 typedef struct {
   CinnamonApp *app;
   MetaWorkspace *active_workspace;
@@ -817,6 +846,7 @@ _cinnamon_app_set_entry (CinnamonApp       *app,
   if (entry != NULL)
     {
       app->info = g_object_ref (gmenu_tree_entry_get_app_info (entry));
+      app->is_flatpak = app->info && gmenu_desktopappinfo_get_is_flatpak (app->info);
     }
 }
 
@@ -1005,7 +1035,7 @@ _cinnamon_app_get_desktop_path (CinnamonApp *app)
 {
   if (app->entry)
     {
-      return g_desktop_app_info_get_filename (app->info);
+      return gmenu_desktopappinfo_get_filename (app->info);
     }
 
   return NULL;
@@ -1073,20 +1103,20 @@ _gather_pid_callback (GDesktopAppInfo   *gapp,
                                                app);
 }
 
-/**
- * cinnamon_app_launch:
- * @timestamp: Event timestamp, or 0 for current event timestamp
- * @uris: (element-type utf8): List of uris to pass to application
- * @workspace: Start on this workspace, or -1 for default
- * @startup_id: (out): Returned startup notification ID, or %NULL if none
- * @error: A #GError
- */
-gboolean
-cinnamon_app_launch (CinnamonApp     *app,
+static void
+apply_discrete_gpu_env (GAppLaunchContext *context)
+{
+  g_app_launch_context_setenv (context, "__NV_PRIME_RENDER_OFFLOAD", "1");
+  g_app_launch_context_setenv (context, "__GLX_VENDOR_LIBRARY_NAME", "nvidia");
+}
+
+static gboolean
+real_app_launch (CinnamonApp   *app,
                   guint         timestamp,
                   GList        *uris,
                   int           workspace,
                   char        **startup_id,
+                  gboolean      offload,
                   GError      **error)
 {
   GdkAppLaunchContext *context;
@@ -1124,7 +1154,13 @@ cinnamon_app_launch (CinnamonApp     *app,
   gdk_app_launch_context_set_timestamp (context, timestamp);
   gdk_app_launch_context_set_desktop (context, workspace);
 
-  ret = g_desktop_app_info_launch_uris_as_manager (app->info, uris,
+  if (offload)
+    {
+      apply_discrete_gpu_env (G_APP_LAUNCH_CONTEXT (context));
+      g_debug ("Offloading '%s' to discrete gpu.", cinnamon_app_get_name (app));
+    }
+
+  ret = gmenu_desktopappinfo_launch_uris_as_manager (app->info, uris,
                                                    G_APP_LAUNCH_CONTEXT (context),
                                                    G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD | G_SPAWN_STDOUT_TO_DEV_NULL  | G_SPAWN_STDERR_TO_DEV_NULL,
                                                    NULL, NULL,
@@ -1135,13 +1171,65 @@ cinnamon_app_launch (CinnamonApp     *app,
   return ret;
 }
 
+/**
+ * cinnamon_app_launch:
+ * @timestamp: Event timestamp, or 0 for current event timestamp
+ * @uris: (element-type utf8): List of uris to pass to application
+ * @workspace: Start on this workspace, or -1 for default
+ * @startup_id: (out): Returned startup notification ID, or %NULL if none
+ * @error: A #GError
+ */
+gboolean
+cinnamon_app_launch (CinnamonApp     *app,
+                     guint            timestamp,
+                     GList           *uris,
+                     int              workspace,
+                     char           **startup_id,
+                     GError         **error)
+{
+  return real_app_launch (app,
+                          timestamp,
+                          uris,
+                          workspace,
+                          startup_id,
+                          FALSE,
+                          error);
+}
+
+/**
+ * cinnamon_app_launch_offloaded:
+ * @timestamp: Event timestamp, or 0 for current event timestamp
+ * @uris: (element-type utf8): List of uris to pass to application
+ * @workspace: Start on this workspace, or -1 for default
+ * @startup_id: (out): Returned startup notification ID, or %NULL if none
+ * @error: A #GError
+ *
+ * Launch an application using the dedicated gpu (if available)
+ */
+gboolean
+cinnamon_app_launch_offloaded (CinnamonApp     *app,
+                               guint            timestamp,
+                               GList           *uris,
+                               int              workspace,
+                               char           **startup_id,
+                               GError         **error)
+{
+  return real_app_launch (app,
+                          timestamp,
+                          uris,
+                          workspace,
+                          startup_id,
+                          TRUE,
+                          error);
+}
+
 /**
  * cinnamon_app_get_app_info:
  * @app: a #CinnamonApp
  *
- * Returns: (transfer none): The #GDesktopAppInfo for this app, or %NULL if backed by a window
+ * Returns: (transfer none): The #GMenuDesktopAppInfo for this app, or %NULL if backed by a window
  */
-GDesktopAppInfo *
+GMenuDesktopAppInfo *
 cinnamon_app_get_app_info (CinnamonApp *app)
 {
   return app->info;
